from typing import List, Tuple


class Result:
    updates = 0
    inserts = 0
    deletes = 0
    names = []

    def __str__(self) -> str:
        return f"Result: {self.inserts} inserts, {self.updates} updates, {self.deletes} deletes across {len(self.names)} syncs"

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other) -> bool:
        if not issubclass(type(other), Result):
            return NotImplemented

        return (self.inserts, self.updates, self.deletes) == (
            other.inserts,
            other.updates,
            other.deletes,
        )


class SyncResult(Result):
    def __init__(self, results: List[Tuple[str, int, int, int]] = []):
        super().__init__()
        for result in results:
            self.names.append(result[0])
            self.inserts += result[1] or 0
            self.updates += result[2] or 0
            self.deletes += result[3] or 0



from uuid import uuid4
import polars as pl
import pandas as pd
import sqlalchemy as sqla
import pyodbc
from .interface import (
    register_source,
    SourceInterface,
    SourceInterfaceConfiguration,
    SourceInterfaceConnectionSettings,
)


class SqlserverConfiguration(SourceInterfaceConfiguration):
    """The configuration class used for SQLServer sources"""

    table: str = None
    query: str = None
    soft_delete: bool = False
    deleted_field: str = None
    deleted_value: str = "'true'"
    unique_columns: list[str] = None
    skip_columns: list[str] = None

    def __init__(
        self,
        table: str = None,
        query: str = None,
        soft_delete: bool = False,
        deleted_field: str = None,
        deleted_value: str = "true",
        uniqueColumns: list[str] = [],
        skipColumns: list[str] = [],
    ):
        super().__init__()
        self.table = table
        self.query = query
        self.soft_delete = soft_delete
        self.deleted_field = deleted_field
        self.deleted_value = deleted_value
        self.unique_columns = uniqueColumns
        self.skip_columns = skipColumns


class SqlserverConnectionSettings(SourceInterfaceConnectionSettings):
    """The connection configuration class used for SQLServer sources"""

    data: pl.DataFrame
    connection_string: str
    query: str = None
    table: str = None

    def __init__(self, settings: dict):
        if settings.get("connection_string", False):
            self.connection_string = settings["connection_string"]
            if "driver" not in self.connection_string:
                if len(pyodbc.drivers()) == 0:
                    raise pyodbc.DatabaseError(
                        "No ODBC drivers found for SQL Server/pyodbc. "
                        "Please make sure at least one is installed"
                    )
                print("Found other ODBDC driver than the one provided, trying...\n")
                for driver in pyodbc.drivers():
                    print(f"Trying driver: {driver}")
                    if "SQL Server" in driver:
                        if "?" not in self.connection_string:
                            self.connection_string += f"?driver={driver}"
                        else:
                            self.connection_string += f"&driver={driver}"
                        break

                if "driver" not in self.connection_string:
                    raise pyodbc.DatabaseError(
                        "No working ODBC drivers found for SQL Server/pyodbc. "
                        "Please make sure at least one is installed"
                    )

            return

        self.connection_string = "mssql://"
        f"{settings['username']}:{settings['password']}"
        f"@{settings['host']}:{settings['port']}/{settings['database']}"


@register_source("sqlserver", SqlserverConfiguration, SqlserverConnectionSettings)
class SqlserverSource(SourceInterface):
    ConnectionSettingsClass = SqlserverConnectionSettings
    SourceConfigClass = SqlserverConfiguration
    connection: pyodbc.Connection = None

    """ A source for SqlServer data """

    def _configure(self):
        pass

    def _connect(self):
        try:
            engine = sqla.create_engine(
                self.connection_settings.connection_string)
        except ModuleNotFoundError:
            engine = sqla.create_engine(
                self.connection_settings.connection_string.replace(
                    "mysql://", "mysql+pymysql://"
                ))
        self.connection = engine.connect()

    def _disconnect(self):
        self.connection.commit()
        self.connection.close()

    def _query(
        self, params=None, customQuery: str = None, returnData: bool = True
    ) -> pl.DataFrame:
        query = customQuery

        if query is None:
            if self.source_configuration.query is not None:
                query = self.source_configuration.query

            if query is None:
                if self.source_configuration.table is None:
                    raise Exception("No query or table specified")

                query = f"SELECT * FROM [{self.source_configuration.table}]"

        if returnData:
            pdd = pd.read_sql_query(sql=query, con=self.connection)
            return pl.from_pandas(pdd)

        if isinstance(query, str):
            query = sqla.text(query)
        self.connection.execute(query)

    def _insert(
        self,
        data: pl.DataFrame,
        table: str = None,
        tempDB: bool = False,
    ):
        # Check for binary tables
        data_columns = data.get_columns()
        for column in data_columns:
            if column.dtype == pl.Binary:
                data = data.with_columns(
                    pl.col(column.name).map_elements(
                        lambda x: bytes(x), return_dtype=pl.Binary
                    )
                )

        if table is None:
            table = self.source_configuration.table

        pand_df = data.to_pandas()
        try:
            pand_df.to_sql(
                table,
                if_exists="append" if not tempDB else "replace",
                index=False,
                con=self.connection,
            )
        except ModuleNotFoundError:
            data.write_database(
                table,
                connection=self.connection,
                if_exists="append" if not tempDB else "replace",
            )

        return len(data)

    def insert(self, data: pl.DataFrame):
        if len(data) == 0:
            return 0

        return self._insert(data)

    def update(self, data: pl.DataFrame):
        self._validate_unique_columns()
        if len(data) == 0:
            return 0

        tempDB = self._get_temp_table_name(self.source_configuration.table)

        # Insert data into temporary table
        self._insert(data, table=tempDB, tempDB=True)

        columns_to_update = [
            column_name
            for column_name in data.columns
            if column_name not in self.source_configuration.unique_columns
            and column_name not in self.source_configuration.skip_columns
        ]
        field_spec = ", ".join(
            (
                f"TDEST.[{column_name}] = TTEMP.[{column_name}]"
                for column_name in columns_to_update
            )
        )

        on_clause = " AND ".join(
            (
                f"TDEST.[{column_name}] = TTEMP.[{column_name}]"
                for column_name in self.source_configuration.unique_columns
            )
        )

        query = f"""
            UPDATE TDEST
            SET {field_spec}
            FROM {self.source_configuration.table} AS TDEST
            INNER JOIN {tempDB} AS TTEMP ON {on_clause}
        """
        self._query(customQuery=query, returnData=False)
        try:
            self._query(customQuery="DROP TABLE " + tempDB, returnData=False)
        except Exception:
            pass

        return len(data)

    def delete(self, data: pl.DataFrame):
        self._validate_unique_columns()
        if len(data) == 0:
            return 0

        tempDB = self._get_temp_table_name(self.source_configuration.table)
        self._insert(data, table=tempDB, tempDB=True)

        where_clause = " AND ".join(
            (
                f"TTEMP.[{column_name}] = TDEST.[{column_name}]"
                for column_name in self.source_configuration.unique_columns
            )
        )

        if self.source_configuration.soft_delete:
            if (
                self.source_configuration.deleted_field is None
                or self.source_configuration.deleted_value is None
            ):
                raise Exception(
                    "Deleted field and value must be specified when using soft delete"
                )

            query = f"""
                UPDATE TDEST
                SET {self.source_configuration.deleted_field} = {self.source_configuration.deleted_value}
                FROM {self.source_configuration.table} AS TDEST
                WHERE EXISTS (
                    SELECT 1/0
                    FROM {tempDB} AS TTEMP
                    WHERE {where_clause}
                ) 
            """
        else:
            query = f"""
                DELETE TDEST
                FROM {self.source_configuration.table} AS TDEST
                WHERE EXISTS (
                    SELECT 1/0
                    FROM {tempDB} AS TTEMP
                    WHERE {where_clause}
                ) 
            """

        self._query(customQuery=query, returnData=False)

        try:
            self._query(customQuery="DROP TABLE " + tempDB, returnData=False)
        except Exception:
            pass

        return len(data)

    def _validate_unique_columns(self):
        if not self.source_configuration.unique_columns:
            raise ValueError(
                "Unique columns must be specified for SQLServer source when used as destination"
            )

    def _get_temp_table_name(self, table_name: str):
        return f"##{table_name}_{str(uuid4()).replace('-', '')}_temp"



import polars as pl
import pandas as pd
from pyarrow.lib import ArrowInvalid
from .interface import (
    register_source,
    SourceInterface,
    SourceInterfaceConfiguration,
    SourceInterfaceConnectionSettings,
)
import requests


class RestResponse:
    length: str
    items: str

    def __init__(self, length: str, items: str):
        self.length = length
        self.items = items


class BodyOptions:
    accepts_multiple: bool = False
    object_path: str = ""

    def __init__(self, accepts_multiple: bool = False, object_path: str = ""):
        self.accepts_multiple = accepts_multiple
        self.object_path = object_path


class PaginationSettings:
    enabled: bool = False
    pageSize: int = 10
    startPage: int = 0
    pageSizeQuery: str = "pageSize"
    pageQuery: str = "page"
    totalQuery: str = "page.count"
    queryIn: str = "query"
    totalQueryIn: str = "body"

    def __init__(self, **kwargs):
        self.enabled = kwargs.get("enabled", False)
        self.pageSizeQuery = kwargs.get("pageSizeQuery", "pageSize")
        self.pageSize = kwargs.get("pageSize", 10)
        self.startPage = kwargs.get("startPage", 0)
        self.pageQuery = kwargs.get("pageQuery", "page")
        self.totalQuery = kwargs.get("totalQuery", "count")
        self.queryIn = kwargs.get("queryIn", "query")


class RestRequest:
    path: str = None
    query: dict = {}
    method: str = "GET"
    pagination: PaginationSettings = None
    body_type: str = "application/json"
    body = None
    body_options: BodyOptions = None
    return_type: str = "application/json"
    response: RestResponse = None

    def __init__(self, path: str = None, query: dict = {}, method: str = "GET", body_type: str = "application/json", body=None, body_options: dict = None, return_type: str = "application/json", response: dict = None, pagination: dict = None):
        self.path = path
        self.method = method
        self.body_type = body_type
        self.body = body
        self.return_type = return_type
        if body_options is not None:
            self.body_options = BodyOptions(**body_options)
        if pagination is not None:
            self.pagination = PaginationSettings(**pagination)
        self.response = response
        self.query = query


class RestSourceConfiguration(SourceInterfaceConfiguration):
    """The configuration class used for MySQL sources"""
    listRequest: RestRequest
    createRequest: RestRequest
    updateRequest: RestRequest
    deleteRequest: RestRequest

    def __init__(self, listRequest: dict = None, createRequest: dict = None, updateRequest: dict = None, deleteRequest: dict = None):
        super().__init__()
        self.listRequest = RestRequest(**listRequest)
        self.createRequest = RestRequest(
            **createRequest if createRequest is not None else {})
        self.updateRequest = RestRequest(
            **updateRequest if updateRequest is not None else {})
        self.deleteRequest = RestRequest(
            **deleteRequest if deleteRequest is not None else {})


class Oauth2Settings:
    def __init__(self, flow: str = "", token_url: str = "", authorization_url: str = "", token_body: dict = {}, token_path: str = "access_token", valid_to_path: str = "expires_on"):
        self.flow = flow
        self.token_url = token_url
        self.authorization_url = authorization_url
        self.token_body = token_body
        self.token_path = token_path
        self.valid_to_path = valid_to_path


class RestAuthentication:
    basic: bool = False
    basic_user: str = None
    basic_pass: str = None
    bearer: bool = False
    bearer_prefix: str = "Bearer"
    bearer_token: str = None
    oauth2: bool = False
    oauth2_settings: Oauth2Settings = None

    # def __init__(self, basic: bool = False, basic_user: str = None, basic_pass: str = None, bearer: bool = False, bearer_prefix: str = "Bearer", bearer_token: str = None, soft_delete: bool = False, delete_field: str = "deleted", deleted_value = None):

    def __init__(self, **kvargs):
        self.basic = kvargs.get("basic", False)
        self.basic_user = kvargs.get("basic_user", None)
        self.basic_pass = kvargs.get("basic_pass", None)
        self.bearer = kvargs.get("bearer", False)
        self.bearer_prefix = kvargs.get("bearer_prefix", "Bearer")
        self.bearer_token = kvargs.get("bearer_token", None)
        self.oauth2 = kvargs.get("oauth2", False)
        self.oauth2_settings = kvargs.get("oauth2_settings", {})
        if self.oauth2:
            self.oauth2_settings = Oauth2Settings(**kvargs["oauth2_settings"])


class RestSourceConnectionSettings(SourceInterfaceConnectionSettings):
    """The connection configuration class used for MySQL sources"""

    base_url: str
    authentication: RestAuthentication = None
    ignore_certificates: bool = False
    soft_delete: bool = False
    deleted_field: str = "deleted"
    deleted_value = None
    client = None

    def __init__(self, settings: dict):
        if settings.get("base_url", None) is not None:
            self.base_url = settings["base_url"]

        if settings.get("authentication", None) is not None:
            self.authentication = RestAuthentication(
                **settings["authentication"])

        if settings.get("ignore_certificates", None) is not None:
            self.ignore_certificates = settings["ignore_certificates"]

        self.soft_delete = settings.get("soft_delete", False)
        self.deleted_field = settings.get("deleted_field", "deleted")
        self.deleted_value = settings.get("deleted_value", None)


@register_source("rest", RestSourceConfiguration, RestSourceConnectionSettings)
class RestSource(SourceInterface):
    ConnectionSettingsClass = RestSourceConnectionSettings
    SourceConfigClass = RestSourceConfiguration
    source_configuration: RestSourceConfiguration
    connection_settings: RestSourceConnectionSettings
    client = None
    authValidTo = None

    """ A source for MySQL data """

    def _configure(self):
        if self.client is None:
            self.client = requests.Session()
            if not self.connection_settings.authentication:
                return
            if self.connection_settings.authentication.basic:
                self.client.auth = (self.connection_settings.authentication.basic_user,
                                    self.connection_settings.authentication.basic_pass)

            if self.connection_settings.authentication.bearer:
                self.client.headers.update(
                    {"Authorization": f"{self.connection_settings.authentication.bearer_prefix} {self.connection_settings.authentication.bearer_token}"})

            if self.connection_settings.authentication.oauth2:
                if self.connection_settings.authentication.oauth2_settings.flow == "client_credentials":
                    response = self.client.post(
                        self.connection_settings.authentication.oauth2_settings.token_url,
                        data=self.connection_settings.authentication.oauth2_settings.token_body
                    )
                    if response.status_code != 200:
                        raise Exception(
                            "Failed to authenticate: ", response.text)

                    response_data = response.json()
                    self.client.headers.update(
                        {"Authorization": f"{self.connection_settings.authentication.bearer_prefix} {response_data[self.connection_settings.authentication.oauth2_settings.token_path]}"})

    def _connect(self):
        if self.authValidTo is None:
            return True

        if self.authValidTo < pd.Timestamp.now():
            return self._configure()

    def _disconnect(self):
        pass

    def _query(
        self, params=None, customQuery: str = None, returnData: bool = True
    ) -> pl.DataFrame:
        self._connect()

        lrSettings = self.source_configuration.listRequest

        request = {
            "method": lrSettings.method,
            "url": self.connection_settings.base_url + lrSettings.path,
            "headers": {"Content-Type": lrSettings.body_type},
            "verify": (not self.connection_settings.ignore_certificates),
            "params": {}
        }

        if lrSettings.return_type != "":
            request["headers"]["Accept"] = lrSettings.return_type

        if lrSettings.body_type == "application/json" and lrSettings.body is not None:
            request["json"] = lrSettings.body

        if lrSettings.query is not None:
            request["params"] = lrSettings.query

        else:
            request["data"] = lrSettings.body

        if lrSettings.pagination is not None and lrSettings.pagination.enabled:
            response_data = []
            page = lrSettings.pagination.startPage
            pageSize = lrSettings.pagination.pageSize
            total = 1

            while len(response_data) < total:
                request["params"][lrSettings.pagination.pageQuery] = page
                request["params"][lrSettings.pagination.pageSizeQuery] = pageSize

                response = self.client.request(**request)
                if response.status_code != 200:
                    if len(response_data) > 0:
                        break
                    raise Exception("Failed request")

                if lrSettings.return_type == "application/json":
                    page_response_data = response.json()
                else:
                    raise Exception("Not implemented")

                if lrSettings.pagination.totalQueryIn != "":
                    if lrSettings.pagination.totalQueryIn == "body":
                        tq = page_response_data
                        for key in lrSettings.pagination.totalQuery.split("."):
                            tq = tq[key]

                        if tq > total:
                            total = tq

                    elif lrSettings.pagination.totalQueryIn == "header":
                        total = response.headers[lrSettings.pagination.totalQuery]

                if len(lrSettings.response["items"]) > 0:
                    for key in lrSettings.response["items"].split("."):
                        page_response_data = page_response_data[key]

                if len(response_data) == 0 and len(page_response_data) < pageSize:
                    pageSize = len(page_response_data)

                response_data += page_response_data
                page += 1

        else:
            response = self.client.request(**request)

            if response.status_code != 200:
                raise Exception("Failed request: ", response.text)

            if self.source_configuration.listRequest.return_type == "application/json":
                response_data = response.json()
            else:
                raise Exception("Not implemented")

            if (self.source_configuration.listRequest.response and
                self.source_configuration.listRequest.response["items"] and
                    len(self.source_configuration.listRequest.response["items"]) > 0):
                for key in self.source_configuration.listRequest.response["items"].split("."):
                    response_data = response_data[key]

            if len(response_data) == 0:
                return pl.DataFrame()

        pd_df = pd.json_normalize(response_data)

        try:
            return pl.from_pandas(pd_df)
        except ArrowInvalid:
            for col in pd_df.columns:
                if pd_df[col].dtype == "object":
                    pd_df[col] = pd_df[col].astype("string")

            return pl.from_pandas(pd_df)

    def _insert(
        self, data: pl.DataFrame,
    ):
        self._connect()

        request = {
            "method": self.source_configuration.createRequest.method,
            "url": self.connection_settings.base_url + self.source_configuration.createRequest.path,
            "headers": {"Content-Type": self.source_configuration.createRequest.body_type},
            "verify": (not self.connection_settings.ignore_certificates)
        }

        body_template = {}
        body_path = []

        if self.source_configuration.createRequest.body_options.object_path != "":
            body_template = self.source_configuration.createRequest.body
            body_path = self.source_configuration.createRequest.body_options.object_path.split(
                ".")

        if self.source_configuration.createRequest.body_options.accepts_multiple:
            body = dict(body_template)
            allItems = []
            for item in data.iter_rows(named=True):
                allItems.append(item)

            if len(body_path) == 1:
                body[body_path[0]] = allItems
            elif len(body_path) == 2:
                body[body_path[0]][body_path[1]] = allItems
            elif len(body_path) == 3:
                body[body_path[0]][body_path[1]][body_path[2]] = allItems
            elif len(body_path) == 4:
                body[body_path[0]][body_path[1]
                                   ][body_path[2]][body_path[3]] = allItems
            else:
                body = allItems

            request["json"] = body

            response = self.client.request(**request)
            if response.status_code > 299:
                raise Exception("Insertion failed: ", response.text)

        else:
            for item in data.iter_rows(named=True):
                body = dict(body_template)

                if len(body_path) == 1:
                    if body_path[0] == "0":
                        body = [item]
                    else:
                        body[body_path[0]] = item
                elif len(body_path) == 2:
                    if body_path[1] == "0":
                        body[body_path[0]] = [item]
                    else:
                        body[body_path[0]][body_path[1]] = item
                elif len(body_path) == 3:
                    if body_path[2] == "0":
                        body[body_path[0]][body_path[1]] = [item]
                    else:
                        body[body_path[0]][body_path[1]][body_path[2]] = item
                elif len(body_path) == 4:
                    if body_path[3] == "0":
                        body[body_path[0]][body_path[1]][body_path[2]] = [item]
                    else:
                        body[body_path[0]][body_path[1]
                                           ][body_path[2]][body_path[3]] = item
                else:
                    body = item

                if self.source_configuration.createRequest.body_type == "application/json":
                    request["json"] = body
                else:
                    request["data"] = body

                response = self.client.request(**request)
                if response.status_code > 299:
                    print("Request failed: ", response.text)

    def insert(self, data: pl.DataFrame):
        return self._insert(data)
        raise NotImplementedError("Insert not implemented")

    def update(self, data: pl.DataFrame):
        raise NotImplementedError("Update not implemented")

    def delete(self, data: pl.DataFrame):
        if self.connection_settings.soft_delete != True and self.source_configuration.deleteRequest.path == "":
            print("Delete not configured, skipping")
            return 0
        return 0



import polars as pl
from .interface import TransformerInterface, register_transformer_class


def dn_to_samaccountname(dn):
    return dn.split(",")[0].split("=")[1]


@register_transformer_class("misc")
class MiscTransformer(TransformerInterface):
    @staticmethod
    def sam_from_dn(
        data: pl.DataFrame, inField: str, outField: str = None
    ) -> pl.DataFrame:
        """Get the SAM account name from a distinguished name

        Args:
            data (pl.DataFrame): The data to be modified
            inField (str): The field to work on
            outField (str, optional): The field to add/replace. Defaults to None.

        Returns:
            pl.DataFrame: Dataframe with the modified column
        """

        data = data.with_columns(
            data[inField]
            .str.splitn(",", 2)
            .struct.unnest()["field_0"]
            .str.splitn("=", 2)
            .struct.unnest()["field_1"]
            .alias(outField if outField is not None else inField)
        )

        return data



import os
import yaml
import json
from typing import Any, List, Dict
from dataclasses import dataclass
import copy
import polars as pl
from .sources.interface import (
    ColumnDefinition,
    SourceInterfaceConfiguration,
    SourceInterfaceConnectionSettings,
)
from .transformers.interface import TransformerConfiguration


class SourceSettings:
    """The configuration class used for datasources"""

    name: str
    source_type: str
    connection: SourceInterfaceConnectionSettings
    config: SourceInterfaceConfiguration

    def __init__(
        self, name: str, source_type: str, connection: dict, config: dict
    ) -> None:
        self.name, self.source_type = name, source_type
        source_class = __import__(f"beetl.sources.{source_type.lower()}")

        self.connection = getattr(
            source_class, f"{source_type}SourceConnectionSettings"
        )(**connection)

        self.config = getattr(
            source_class, f"{source_type}SourceConfig")(**config)


@dataclass
class ChangeDetectionConfig:
    threshold: int = -1
    stopSync: bool = False
    webhook: str = None
    webhookHeaders: Dict[str, str] = None
    webhookPayload: Dict[str, str] = None


@dataclass
class ComparisonColumn:
    name: str
    type: pl.DataType
    unique: bool = False

    def __init__(self, name: str, type: Any, unique: bool = False) -> None:
        self.name, self.unique = name, unique
        self.type = getattr(pl, type)


@dataclass
class SyncConfiguration:
    """The configuration for a single sync between two sources"""

    source: SourceSettings
    sourceConfig: SourceInterfaceConfiguration
    destination: SourceSettings
    destinationConfig: SourceInterfaceConfiguration
    comparisonColumns: List[ComparisonColumn]
    name: str = ""
    changeDetection: ChangeDetectionConfig = None
    sourceTransformers: TransformerConfiguration = None
    destinationTransformers: TransformerConfiguration = None
    insertionTransformers: List[TransformerConfiguration] = None
    deletionTransformers: List[TransformerConfiguration] = None

    def __post_init__(self) -> None:
        self.source.config = self.sourceConfig
        self.destination.config = self.destinationConfig


class BeetlConfig:
    """The configuration class for BeETL"""

    version: str
    sources: Dict[str, SourceSettings]
    sync_list: List[SyncConfiguration]

    def __init__(self, config: dict) -> None:
        # Set version to keep track of changes
        # in an effort to keep backwards compatibility

        # Get the class corresponding to the configuration version
        module = __import__(self.__module__, fromlist=[""])
        config_class = getattr(
            module, f'BeetlConfig{config.get("configVersion", "V1")}'
        )

        # Transform this class into the versioned configuration class
        self.__class__ = config_class
        self.__dict__ = config_class(config).__dict__
        self.version = config.get("configVersion", "V1")

    @classmethod
    def from_yaml(cls, yamlData: str) -> "BeetlConfig":
        """Import configuration from a YAML string

        Args:
            yaml (str): The YAML string containing the configuration

        Returns:
            BeetlConfig: Beetl Configuration
        """
        config = yaml.safe_load(yamlData)
        return cls(config)

    @classmethod
    def from_yaml_file(cls, path: str, encoding: str = "utf-8") -> "BeetlConfig":
        """Import configuration from YAML file

        Args:
            path (str): The path to the configuration file
            encoding (str, optional): The encoding of the file. Defaults to 'utf-8'.

        Raises:
            Exception: File not found

        Returns:
            BeetlConfig: Beetl Configuration
        """
        try:
            with open(path, "r", encoding=encoding) as file:
                config = yaml.safe_load(file)
                return cls(config)
        except FileNotFoundError as e:
            raise Exception(
                f"The configuration file was not found at the path specified ({path})."
            ) from e

    @classmethod
    def from_json(cls, jsonData: str) -> "BeetlConfig":
        """Import configuration from a JSON string

        Args:
            jsonData (str): The JSON string containing the configuration

        Returns:
            BeetlConfig: Beetl Configuration
        """
        config = json.loads(jsonData)
        return cls(config)

    @classmethod
    def from_json_file(cls, path: str, encoding: str = "utf-8") -> "BeetlConfig":
        """Import configuration from YAML file

        Args:
            path (str): The path to the configuration file
            encoding (str, optional): The encoding of the file. Defaults to 'utf-8'.

        Raises:
            Exception: File not found

        Returns:
            BeetlConfig: Beetl Configuration
        """
        try:
            with open(path, "r", encoding=encoding) as file:
                config = json.load(file)
                return cls(config)
        except FileNotFoundError as e:
            raise Exception(
                f"The configuration file was not found at the path specified ({path})."
            ) from e


class BeetlConfigV1(BeetlConfig):
    """Version 1 of the Beetl Configuration Interface"""

    def __init__(self, config: dict):
        self.sources, self.sync_list = {}, []

        if len(config.get("sync", "")) == 0:
            raise Exception(
                "The configuration file is missing the 'sync' section.")

        if len(config.get("sources", "")) == 0:
            if config.get("sourcesFromEnv", "") in [None, ""]:
                raise Exception(
                    "The configuration file is missing the 'sources' section."
                    "If you are using environment variables to configure your sources, "
                    "please set the 'sourcesFromEnv' property to the name"
                    "of the environment variable."
                )

            sourcesFromEnv = config.pop("sourcesFromEnv")
            envJson = os.environ.get(sourcesFromEnv)
            try:
                envConfig = json.loads(envJson)
            except Exception as e:
                raise Exception(
                    "The environment variable specified in 'sourcesFromEnv' "
                    "is not a valid JSON string or does not exist."
                ) from e

            config["sources"] = envConfig

        for source in config["sources"]:
            name = source.pop("name")
            source_type = source.pop("type")
            source_module = __import__(
                ".".join(
                    self.__module__.split(".")[0:-1]
                    + [f"sources.{source_type.lower()}"]
                ),
                fromlist=[""],
            )
            source_class = getattr(source_module, f"{source_type}Source")
            self.sources[name] = source_class(**source)

        for sync in config["sync"]:
            if not self.sources.get(sync["source"], False) or not self.sources.get(
                sync["destination"], False
            ):
                raise Exception(
                    "One of the source/destination names in "
                    "the sync section does not match a source name "
                    "in the sources section."
                    "Please check your configuration.",
                )

            tmpSource = copy.deepcopy(self.sources[sync["source"]])
            tmpSource.set_sourceconfig(sync["sourceConfig"])
            tmpDestination = copy.deepcopy(self.sources[sync["destination"]])
            tmpDestination.set_sourceconfig(sync["destinationConfig"])

            comparisonColumnsConf = sync.get("comparisonColumns", None)
            if not comparisonColumnsConf:
                raise Exception(
                    "The sync configuration is missing the 'comparisonColumns' key."
                )
            if type(comparisonColumnsConf) is list:
                try:
                    comparisonColumns = [
                        ComparisonColumn(**args) for args in sync["comparisonColumns"]
                    ]
                except TypeError as e:
                    raise Exception(
                        "When passing the comparisonColumns as a list it must be a list of dictionaries containing the mandatory keys 'name', 'type', and optionally 'unique'."
                    ) from e
            elif type(comparisonColumnsConf) is dict:
                try:
                    comparisonColumns = []
                    unique_column_name, _ = list(
                        comparisonColumnsConf.items())[0]
                    for key, value in comparisonColumnsConf.items():
                        comparisonColumns.append(ComparisonColumn(
                            key, value, key == unique_column_name))

                except Exception as e:
                    raise Exception(
                        "The comparisonColumns must be a list of dictionaries containing the mandatory keys 'name', 'type', and optionally 'unique'."
                    ) from e
            else:
                raise Exception(
                    "The comparisonColumns must be a list of dictionaries containing the mandatory keys 'name', 'type', and optionally 'unique'. Or a dictionary with the column names as key and types as values, where the first key is treated as the unique column."
                )

            syncConfig = SyncConfiguration(
                name=sync.get("name", ""),
                source=tmpSource,
                sourceConfig=sync["sourceConfig"],
                destination=tmpDestination,
                destinationConfig=sync["destinationConfig"],
                comparisonColumns=comparisonColumns,
            )

            tmpSource = tmpDestination = None

            if sync.get("sourceTransformer", None) is not None:
                syncConfig.sourceTransformer = TransformerConfiguration(
                    "source", sync["sourceTransformer"], {}
                )

            if sync.get("sourceTransformers", None) is not None:
                syncConfig.sourceTransformers = []
                for transformer in sync["sourceTransformers"]:
                    syncConfig.sourceTransformers.append(
                        TransformerConfiguration(
                            transformer.get("transformer"),
                            transformer.get("config", None),
                        )
                    )

            if sync.get("destinationTransformers", None) is not None:
                syncConfig.destinationTransformers = []
                for transformer in sync["destinationTransformers"]:
                    syncConfig.destinationTransformers.append(
                        TransformerConfiguration(
                            transformer.get("transformer"),
                            transformer.get("config", None),
                        )
                    )

            if sync.get("insertionTransformers", None) is not None:
                syncConfig.insertionTransformers = []
                for transformer in sync["insertionTransformers"]:
                    syncConfig.insertionTransformers.append(
                        TransformerConfiguration(
                            transformer.get("transformer"),
                            transformer.get("config", None),
                            transformer.get("include_sync", False),
                        )
                    )

            if sync.get("deletionTransformers", None) is not None:
                syncConfig.deletionTransformers = []
                for transformer in sync["deletionTransformers"]:
                    syncConfig.deletionTransformers.append(
                        TransformerConfiguration(
                            transformer.get("transformer"),
                            transformer.get("config", None),
                            transformer.get("include_sync", False),
                        )
                    )

            self.sync_list.append(syncConfig)



import polars as pl
from .interface import TransformerInterface, register_transformer_class


@register_transformer_class("regex")
class RegexTransformer(TransformerInterface):
    @staticmethod
    def replace(
        data: pl.DataFrame,
        query: str,
        replace: str,
        inField: str,
        outField: str = None,
        maxN: int = -1,
    ) -> pl.DataFrame:
        """Run a regex replace operation on the given field

        Args:
            data (pl.DataFrame): The data to be modified
            query (str): The regex query to run
            replace (str): The replacement string
            inField (str): The field to work on
            outField (str, optional): The field to add/replace. Defaults to None.

        Returns:
            pl.DataFrame: Dataframe with the modified column
        """

        data = data.with_columns(
            data[inField]
            .str.replace(query, replace, n=maxN)
            .alias(outField if outField is not None else inField)
        )

        return data



from dataclasses import dataclass
import polars as pl
from typing import List
import concurrent.futures

CASTABLE = (
    pl.Int8,
    pl.Int16,
    pl.Int32,
    pl.Int64,
    pl.UInt8,
    pl.UInt16,
    pl.UInt32,
    pl.UInt64,
    pl.Float32,
    pl.Float64,
    pl.Boolean,
    pl.Utf8,
    pl.Object,
    # Does not quite work
    # pl.Binary,
)


class RequestThreader:
    threads: int = None
    executor: concurrent.futures.ThreadPoolExecutor = None

    def __init__(self, threads: int = 10):
        print("Starting threader...")
        self.threads = threads

    def __enter__(self):
        print("Threading the needle...")
        self.executor = concurrent.futures.ThreadPoolExecutor(
            max_workers=self.threads)

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        print("Waiting for threads to finish...")
        self.executor.shutdown(wait=True)

    def submit(self, func, *args, **kwargs):
        print("Submitting single thread...")
        return self.executor.submit(func, *args, **kwargs)

    def submitAndWait(self, func, kwarg_list: list):
        print("Submitting a list of threads...")
        for result in self.executor.map(lambda fn: func(**fn), kwarg_list):
            yield result


def register_source(name: str, configuration: type, connection_settings: type):
    def wrapper(cls: type):
        Sources.sources[name] = {
            "class": cls,
            "configuration": configuration,
            "connection_settings": connection_settings,
        }
        return cls

    return wrapper


class Sources:
    sources: dict = {}


@dataclass
class ColumnDefinition:
    """The definition of a column in a dataset.
    Name: Name of column
    Type: Polars data type
    Unique: Whether it is unique
    Skip Update: Whether to skip updating this field when inserting/updating source
    Custom Options: Custom options for some providers
    """

    name: str
    type: pl.DataType
    unique: bool = False
    skip_update: bool = False
    custom_options: dict = None

    def __post_init__(self) -> None:
        """Get the actual data type from Polars"""
        self.type = getattr(pl, self.type)


@dataclass
class SourceInterfaceConfiguration:
    """The configuration class used for data sources, abstract"""

    def __init__(self):
        pass


@dataclass
class SourceInterfaceConnectionSettings:
    """The connection configuration class used for data sources, abstract"""

    pass


class SourceInterface:
    ConnectionSettingsClass = SourceInterfaceConnectionSettings
    SourceConfigClass = SourceInterfaceConfiguration

    """ Abstract interface for a connection to a data source """
    connection = None
    connection_settings = None
    source_configuration = None

    def __init__(self, connection: dict, config: dict = None) -> None:
        """Initiates a source class

        Args:
            config (dict): Coniguration for the source data (columns, IDs, etc.)
            connection (dict):
                Configuration for the source connection (paths, credentials, etc.)
        """
        try:
            self.connection_settings = self.ConnectionSettingsClass(
                **connection)
            self.source_configuration = (
                self.SourceConfigClass(**config) if config else None
            )
        except Exception as e:
            raise Exception(
                "Invalid connection settings for source: " + str(e))

        self._configure()

    def set_sourceconfig(self, config: dict) -> None:
        self.source_configuration = self.SourceConfigClass(**config)

    def __enter__(self):
        self._connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._disconnect()

    def _configure(self):
        """
        Method responsible for configuring the connection to
        the dataset. Should be overridden in child class.
        Called during init.
        """
        raise NotImplementedError

    def _connect(self):
        """
        Method responsible for connecting to the dataset.
        Should be overridden in child class. Called during init.
        """
        raise NotImplementedError

    def _disconnect(self):
        """
        Method responsible for disconnecting from the dataset.
        Should be overridden in child class. Called during exit.
        """
        raise NotImplementedError

    def connect(self):
        self._connect()

    def disconnect(self):
        self._disconnect()

    def query(self, params=None):
        """Formats the query according to the column specification

        Args:
            df (DataFrame): The data from the data source
        """
        return self._query(params)

    def _query(self, params=None) -> pl.DataFrame:
        """Run a query on the source and return the results

        Args:
            params (mixed, optional): Parameters for the query. Defaults to None.

        Returns:
            pl.DataFrame: The results of the query
        """
        raise NotImplementedError

    def insert(self, data: pl.DataFrame) -> None:
        """Insert data into the source

        Args:
            data (pl.DataFrame): The data to insert
        """
        raise NotImplementedError

    def update(self, data: pl.DataFrame) -> None:
        """Update data in the source

        Args:
            data (pl.DataFrame): The data to update
        """
        raise NotImplementedError

    def delete(self, data: pl.DataFrame) -> None:
        """Delete data from the source

        Args:
            data (pl.DataFrame): The data to delete
        """
        raise NotImplementedError



from typing import List
from polars import DataFrame, Object
from .interface import (
    register_source,
    SourceInterface,
    SourceInterfaceConfiguration,
    SourceInterfaceConnectionSettings,
)
from pymongo import MongoClient, UpdateOne, DeleteOne


class MongoDBSourceConfiguration(SourceInterfaceConfiguration):
    """The configuration class used for MongoDB sources"""
    collection: str = None
    filter: str = None
    projection: dict = None
    unique_fields: List[str] = None

    def __init__(self, collection: str = None, filter: dict = {}, projection: dict = {}, uniqueFields: List[str] = []):
        super().__init__()
        self.collection = collection
        self.filter = filter
        self.projection = projection
        self.unique_fields = uniqueFields


class MongoDBSourceConnectionSettings(SourceInterfaceConnectionSettings):
    """The connection configuration class used for MongoDB sources"""
    connection_string: str
    query: str = None
    database: str = None

    def get_connection_string_components(self, settings: dict):
        config = {
            "host": settings.get("host", None),
            "port":  settings.get("post", None),
            "username":  settings.get("username", None),
            "password":  settings.get("password", None),
        }
        for _, value in config.items():
            if not value:
                return None

        return config

    def __init__(self, settings: dict):
        self.projection = settings.get("projection", {})

        if not settings.get("database", None):
            raise Exception("Database name in connection settings is required")

        self.database = settings["database"]

        if settings.get("connection_string", False):
            self.connection_string = settings["connection_string"]
        else:
            connection_string_components = self.get_connection_string_components(
                settings)
            if not connection_string_components:
                raise Exception(
                    "Connection string or host, port, username, and password are required")
            self.connection_string = f"mongodb://{connection_string_components['username']}:{connection_string_components['password']}@{connection_string_components['host']}:{connection_string_components['port']}/"


@register_source("mongodb", MongoDBSourceConfiguration, MongoDBSourceConnectionSettings)
class MongodbSource(SourceInterface):
    ConnectionSettingsClass = MongoDBSourceConnectionSettings
    SourceConfigClass = MongoDBSourceConfiguration

    """ A source for MongoDB data """

    def _configure(self):
        pass

    def _connect(self):
        pass

    def _disconnect(self):
        pass

    def _query(self, params=None) -> DataFrame:
        with MongoClient(self.connection_settings.connection_string) as client:
            db = client[self.connection_settings.database]
            collection = db[self.source_configuration.collection]
            polar = DataFrame(collection.find(
                self.source_configuration.filter, projection=self.source_configuration.projection))
            if "_id" in polar.columns and polar["_id"].dtype == Object:
                polar = polar.with_columns(
                    polar["_id"].map_elements(lambda oid: str(oid)))
            return polar

    def insert(self, data: DataFrame) -> int:
        with MongoClient(self.connection_settings.connection_string) as client:
            db = client[self.connection_settings.database]
            collection = db[self.source_configuration.collection]
            result = collection.insert_many(data.to_dicts())

        return len(result.inserted_ids)

    def update(self, data: DataFrame) -> int:
        self._validate_unique_fields()
        updates = []
        for row in data.to_dicts():
            filter = {field_name: row[field_name]
                      for field_name in self.source_configuration.unique_fields}
            for field_name in self.source_configuration.unique_fields:
                del row[field_name]
            updates.append(UpdateOne(filter, {"$set": row}))

        if not len(updates):
            return 0

        with MongoClient(self.connection_settings.connection_string) as client:
            db = client[self.connection_settings.database]
            collection = db[self.source_configuration.collection]
            result = collection.bulk_write(updates)

        return result.modified_count

    def delete(self, data: DataFrame) -> int:
        self._validate_unique_fields()
        deletes = []
        for row in data.to_dicts():
            filter = {field_name: row[field_name]
                      for field_name in self.source_configuration.unique_fields}
            deletes.append(DeleteOne(filter))

        with MongoClient(self.connection_settings.connection_string) as client:
            db = client[self.connection_settings.database]
            collection = db[self.source_configuration.collection]
            result = collection.bulk_write(deletes)

        return result.deleted_count

    def _validate_unique_fields(self):
        if not self.source_configuration.unique_fields:
            raise ValueError(
                "Unique fields are required for MongoDB when used as a destination")



import uuid
import polars as pl
from typing import List

import psycopg
from .interface import (
    register_source,
    SourceInterface,
    SourceInterfaceConfiguration,
    SourceInterfaceConnectionSettings,
)


class PostgresqlSourceConfiguration(SourceInterfaceConfiguration):
    """The configuration class used for Postgresql sources"""

    unique_columns: List[str] = None
    skip_columns: List[str] = None
    table: str = None
    query: str = None

    def __init__(
        self,
        table: str = None,
        query: str = None,
        uniqueColumns: List[str] = [],
        skipColumns: List[str] = [],
    ):
        super().__init__()
        self.table = table
        self.query = query
        self.unique_columns = uniqueColumns
        self.skip_columns = skipColumns


class PostgresqlSourceConnectionSettings(SourceInterfaceConnectionSettings):
    """The connection configuration class used for Postgresql sources"""

    connection_string: str

    def __init__(self, settings: dict):
        if settings.get("connection_string", False):
            self.connection_string = settings["connection_string"]
            return

        self.connection_string = "postgresql://"
        f"{settings['username']}:{settings['password']}"
        f"@{settings['host']}:{settings['port']}/{settings['database']}"


@register_source(
    "postgresql", PostgresqlSourceConfiguration, PostgresqlSourceConnectionSettings
)
class PostgresqlSource(SourceInterface):
    ConnectionSettingsClass = PostgresqlSourceConnectionSettings
    SourceConfigClass = PostgresqlSourceConfiguration

    """ A source for Postgresql data """

    def _configure(self):
        pass

    def _connect(self):
        pass

    def _disconnect(self):
        pass

    def _query(
        self, params=None, customQuery: str = None, returnData: bool = True
    ) -> pl.DataFrame:
        query = customQuery

        if query is None:
            if self.source_configuration.query is not None:
                query = self.source_configuration.query

            if query is None:
                if self.source_configuration.table is None:
                    raise Exception("No query or table specified")

                query = f"SELECT * FROM {self.source_configuration.table}"

        if returnData:
            return pl.read_database_uri(
                query=query,
                uri=self.connection_settings.connection_string,
            )

        with psycopg.connect(self.connection_settings.connection_string) as connection:
            with connection.cursor() as cursor:
                cursor.execute(query)

    def _insert(
        self, data: pl.DataFrame, table: str = None, connection_string: str = None
    ):
        if table is None:
            table = self.source_configuration.table
        if connection_string is None:
            connection_string = self.connection_settings.connection_string

        connection_string = connection_string.replace(
            "postgresql://", "postgresql+psycopg://")

        data.write_database(
            table, connection_string, if_table_exists="append"
        )

        return len(data)

    def insert(self, data: pl.DataFrame):
        return self._insert(data)

    def update(self, data: pl.DataFrame):
        self._validate_unique_columns()
        temp_table_name = (
            self.source_configuration.table
            + "_udTemp_"
            + str(uuid.uuid4()).replace("-", "")
        )

        try:
            with psycopg.connect(
                self.connection_settings.connection_string
            ) as connection:
                with connection.cursor() as cursor:
                    create_temp_table_with_same_structure_as_destination = f"CREATE TABLE {temp_table_name} AS SELECT * FROM {self.source_configuration.table} where 1=0"
                    cursor.execute(
                        create_temp_table_with_same_structure_as_destination)

            self._insert(data, table=temp_table_name)

            source_table_name = temp_table_name
            destination_table_name = self.source_configuration.table

            columns_to_update = [
                column.name
                for column in data.get_columns()
                if not column.name in self.source_configuration.unique_columns
                and not column.name in self.source_configuration.skip_columns
            ]
            set_values_of_comparison_columns = "SET " + ", ".join(
                (
                    f"{columnName} = {source_table_name}.{columnName}"
                    for columnName in columns_to_update
                )
            )

            where_unique_columns_are_matching = "WHERE " + " AND ".join(
                (
                    f"{destination_table_name}.{columnName} = {source_table_name}.{columnName}"
                    for columnName in self.source_configuration.unique_columns
                )
            )

            query = f"""
                UPDATE {self.source_configuration.table} AS {destination_table_name}
                {set_values_of_comparison_columns}
                FROM {temp_table_name} AS {source_table_name}
                {where_unique_columns_are_matching}
            """

            self._query(customQuery=query, returnData=False)

            return len(data)
        finally:
            with psycopg.connect(
                self.connection_settings.connection_string
            ) as connection:
                with connection.cursor() as cursor:
                    cursor.execute(f"DROP TABLE IF EXISTS {temp_table_name}")

    def delete(self, data: pl.DataFrame):
        self._validate_unique_columns()
        batch_size = 500
        batches = [data]

        if len(data) > batch_size:
            batches = []

            for i in range(0, len(data), batch_size):
                batches.append(data[i: i + batch_size])

        for batch in batches:
            id_clause = " AND ".join(
                (
                    f"{columnName} IN ({','.join([self._quote_if_needed(x) for x in batch[columnName].to_list()])})"
                    for columnName in self.source_configuration.unique_columns
                )
            )

            query = f"""
                DELETE FROM {self.source_configuration.table}
                WHERE {id_clause}
            """

            self._query(customQuery=query, returnData=False)

        return len(data)

    def _quote_if_needed(self, id: any) -> str:
        if isinstance(id, str):
            return f"'{id}'"
        return str(id)

    def _validate_unique_columns(self):
        if not self.source_configuration.unique_columns:
            raise ValueError(
                "Unique columns are required for PostgreSQL when used as a destination"
            )



from typing import List
import polars as pl
from .interface import (
    register_source,
    SourceInterface,
    SourceInterfaceConfiguration,
    SourceInterfaceConnectionSettings,
)


class CsvSourceConfiguration(SourceInterfaceConfiguration):
    """The configuration class used for static sources"""

    pass


class CsvSourceConnectionSettings(SourceInterfaceConnectionSettings):
    """The connection configuration class used for static sources"""

    path: str
    encoding: str = "utf-8"

    def __init__(self, path: str, encoding: str = "utf-8"):
        self.path = path
        self.encoding = encoding


@register_source("csv", CsvSourceConfiguration, CsvSourceConnectionSettings)
class CsvSource(SourceInterface):
    ConnectionSettingsClass = CsvSourceConnectionSettings
    SourceConfigClass = CsvSourceConfiguration

    """ A source for static data """

    def _configure(self):
        pass

    def _connect(self):
        pass

    def _disconnect(self):
        pass

    def _query(self, params=None) -> pl.DataFrame:
        return pl.read_csv(self.connection_settings.path, encoding=self.connection_settings.encoding)

    def insert(self, data: pl.DataFrame):
        print("Inserting data into static source...")
        print(data)
        return len(data)

    def update(self, data: pl.DataFrame):
        print("Updating data in static source...")
        print(data)
        return len(data)

    def delete(self, data: pl.DataFrame):
        print("Deleting data from static source")
        print(data)
        return len(data)



import polars as pl
from .interface import (
    TransformerInterface,
    register_transformer,
    register_transformer_class,
)
from bson import ObjectId


@register_transformer_class("strings")
class StringTransformer(TransformerInterface):
    @staticmethod
    def staticfield(data: pl.DataFrame, field: str, value: str) -> pl.DataFrame:
        """Add a static field to the DataFrame

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to process
            stripChar (str): The character to strip

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        data = data.with_columns(pl.Series(field, [value] * len(data)))
        return data

    @staticmethod
    def set_default(data: pl.DataFrame, inField: str, defaultValue: str) -> pl.DataFrame:
        """Set a default value for a column if the current value is null

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to process
            defaultValue (str): The default value

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        __class__._validate_fields(data.columns, inField)

        data = data.with_columns(data[inField].fill_null(defaultValue))
        return data

    @staticmethod
    def strip(data: pl.DataFrame, inField: str, stripChars: str, outField: str = None) -> pl.DataFrame:
        """Strip all given characters from a column

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to process
            stripChar (str): The character to strip

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        __class__._validate_fields(data.columns, inField)

        data = data.with_columns(data[inField].str.strip(stripChars).alias(
            outField if outField is not None else inField))
        return data

    @staticmethod
    def lowercase(data: pl.DataFrame, inField: str = "", outField: str = "", inOutMap: dict = {}) -> pl.DataFrame:
        """Transform all values in a column to lowercase and insert
            them into another (or the same) column

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to take in
            outField (str): The field to put the result in

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        __class__._validate_fields(data.columns, inField)

        if len(inOutMap) == 0:
            inOutMap = {
                inField: outField if outField is not None and outField != "" else inField}

        for inf, outf in inOutMap.items():
            data = data.with_columns(data[inf].str.to_lowercase().alias(outf))

        return data

    @staticmethod
    def uppercase(data: pl.DataFrame, inField: str = "", outField: str = "", inOutMap: dict = {}) -> pl.DataFrame:
        """Transform all values in a column to uppercase
            and insert them into another (or the same) column

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to take in
            outField (str): The field to put the result in

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        if len(inOutMap) == 0:
            inOutMap = {
                inField: outField if outField is not None and outField != "" else inField}

        for inf, outf in inOutMap.items():
            data = data.with_columns(data[inf].str.to_uppercase().alias(outf))

        return data

    @staticmethod
    def match_contains(data: pl.DataFrame, inField: str, match: str, outField: str = "") -> pl.DataFrame:
        """Match a value in a column and insert a boolean value in another column

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to take in
            matchValue (str): The value to match
            outField (str): The field to put the result in
        """

        return data.with_columns(data[inField].str.contains(match).alias(outField if outField is not None and outField != "" else inField))

    @staticmethod
    def join(
        data: pl.DataFrame, inFields: list, outField: str, separator: str = ""
    ) -> pl.DataFrame:
        """Join multiple columns together

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inFields (list): The fields to join
            outField (str): The field to put the result in
            separator (str, optional): The separator to use. Defaults to ''.

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        nCol = pl.concat_str(
            data[inFields], separator=separator).alias(outField)

        return data.with_columns(nCol.alias(outField))

    @staticmethod
    def join_listfield(
        data: pl.DataFrame, inField: str, outField: str, separator: str = ","
    ) -> pl.DataFrame:
        try:
            if isinstance(type(data[inField].dtype), type(pl.List)):
                data = data.with_columns(
                    data[inField].list.join(separator).alias(outField))
                return data

            data = data.with_columns(
                data[inField].arr.join(separator).alias(outField))
            return data

        except Exception:
            data = data.with_columns(data[inField].cast(
                pl.List).arr.join(separator).alias(outField))

            return data

    @staticmethod
    def split(
        data: pl.DataFrame, inField: str, outFields: list, separator: str = ""
    ) -> pl.DataFrame:
        """Split a column into multiple fields

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to split
            outField (str): The fields to put the result in (in order)
            separator (str, optional): The separator to use. Defaults to ''.

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        amount = len(outFields)
        data = data.with_columns(
            data[inField]
            .str.splitn(separator, amount)
            .struct.unnest()
            .rename({f"field_{i}": fld for i, fld in enumerate(outFields)})
        )

        return data

    @staticmethod
    def split_into_listfield(
        data: pl.DataFrame, inField: str, outField: str = None, separator: str = ""
    ) -> pl.DataFrame:
        """Split a column into list field

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to split
            outField (str): The field to put the result into
            separator (str, optional): The separator to use. Defaults to ''.

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        __class__._validate_fields(data.columns, [inField])

        old_column = data[inField]
        new_column = old_column.str.split(separator).alias(outField or inField)

        data = data.with_columns(new_column)

        return data

    @staticmethod
    def quote(data: pl.DataFrame, inField: str, outField: str = None, quote: str = "'"):
        """Quotes the given column values"""

        tmpData = StringTransformer.staticfield(data, "ccQuoteFld", quote)

        new = StringTransformer.join(
            tmpData, ["ccQuoteFld", inField, "ccQuoteFld"], outField, separator=""
        )

        return new

    @staticmethod
    def replace(data: pl.DataFrame, search: str, replace: str, inField: str, outField: str = None):
        """Replace a section in a string with another section"""

        return data.with_columns(data[inField].str.replace(search, replace).alias(outField if outField is not None else inField))

    @staticmethod
    def replace_all(data: pl.DataFrame, search: str, replace: str, inField: str, outField: str = None):
        """Replace all matching sections in a string with another section"""

        return data.with_columns(data[inField].str.replace_all(search, replace).alias(outField if outField is not None else inField))

    @staticmethod
    def substring(data: pl.DataFrame, inField: str, outField: str = None, start: int = 0, length: int = None):
        """Returns a substring of the given string column"""

        return data.with_columns(data[inField].str.slice(start, length).alias(outField if outField is not None else inField))

    @staticmethod
    def add_prefix(data: pl.DataFrame, inField: str, prefix: str, outField: str = None):
        """Add a prefix to the given column"""

        return data.with_columns(pl.concat_str(pl.Series("ccTempField", [prefix] * len(data)), data[inField]).alias(outField if outField is not None else inField))

    @staticmethod
    def cast(data: pl.DataFrame, inField: str, outField: str = ""):
        """Cast a column to a different type"""

        if outField == "":
            outField = inField

        return data.with_columns(data[inField].cast(pl.Utf8).alias(outField))

    @staticmethod
    def hash(data: pl.DataFrame, inField: str, outField: str = ""):
        """Hash a column"""

        if outField == "":
            outField = inField

        return data.with_columns(data[inField].hash().cast(pl.Utf8).alias(outField))

    @staticmethod
    def to_object_id(data: pl.DataFrame, inField: str) -> pl.DataFrame:
        """Convert the given field to a bson.ObjectId

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to convert

        Returns:
            pl.DataFrame: The resulting DataFrame
        """

        data = data.with_columns(
            data[inField].map_elements(lambda oid: ObjectId(oid)))

        return data
